/*
 * An opener for a file provided by an Ext2-compatible filesystem.
 *
 * Copyright (C) 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <dirent.h>
#include <string.h>

#include <thread>

#include <e2access/path.h>
#include <fsclient/client.h>
#include <systypes/fcntl.h>

#include "ext2_file_accessor.h"
#include "ext2_file_opener.h"



/* Common definitions. */

#define DIRENT_CORE_SIZE (sizeof(struct dirent) - sizeof(((struct dirent *) 0)->d_name))

/* File type conversion. */

static int convert_file_type(int type)
{
  switch (type)
  {
    case EXT2_FT_REG_FILE:  return DT_REG;
    case EXT2_FT_DIR:       return DT_DIR;
    case EXT2_FT_CHRDEV:    return DT_CHR;
    case EXT2_FT_BLKDEV:    return DT_BLK;
    case EXT2_FT_FIFO:      return DT_FIFO;
    case EXT2_FT_SOCK:      return DT_SOCK;
    case EXT2_FT_SYMLINK:   return DT_LNK;
    default:                return DT_UNKNOWN;
  }
}

/* Helper function to ensure alignment in generated entries. */

static int pad_align(int value)
{
  return value + (sizeof(unsigned int) - (value % sizeof(unsigned int)));
}

/* Callback function. */

static int read_directory_entry(struct ext2_dir_entry *dir_entry, int offset,
                                int blocksize, char *buf, void *priv_data)
{
    (void) offset; (void) blocksize; (void) buf;

    struct Ext2FileOpenerDir *dir = reinterpret_cast<struct Ext2FileOpenerDir *>(priv_data);

    dir->entry = dir_entry;
    dir->offset = offset;
    return dir->opener->write_directory_entry(dir);
}

/* Thread payload. */

static void _read_directory(Ext2FileOpener *opener, fileid_t fileid, file_t *writer)
{
    /* Subscribe to space and closure notifications on the pipe. */

    long err = client_set_blocking(writer, NOTIFY_SPACE_AVAILABLE);

    if (err)
    {
        client_close(writer);
        return;
    }

    opener->read_directory(fileid, writer);

    client_close(writer);
}



Ext2FileOpener::~Ext2FileOpener()
{
}

/* Test if a directory is being accessed. */

bool Ext2FileOpener::accessing_directory(const char *path, flags_t flags, fileid_t fileid)
{
    (void) path; (void) flags;
    return _ops->is_directory((ext2_ino_t) fileid);
}

/* Test if a file is being accessed. */

bool Ext2FileOpener::accessing_file(const char *path, flags_t flags, fileid_t fileid)
{
    (void) path; (void) flags;
    return _ops->is_file((ext2_ino_t) fileid);
}

// NOTE: This is mostly the same as the HostFileOpener implementation.

long Ext2FileOpener::get_directory(const char *path, flags_t flags, fileid_t fileid, offset_t *size, l4_cap_idx_t *cap)
{
    (void) path; (void) flags;

    file_t *reader, *writer;

    // NOTE: Might be more appropriate to use lower-level file operations to
    // NOTE: avoid unnecessary mapping of the reader's memory region.

    long err = client_pipe(&reader, &writer);

    if (err)
        return err;

    *size = reader->size;
    *cap = reader->ref;

    /* Discard the reader structure but do not close the reader itself. */

    delete reader;

    /* Spawn a independent thread for reading the directory details and writing
       them to the pipe. */

    std::thread(_read_directory, this, fileid, writer).detach();

    return L4_EOK;
}

/* Thread payload helper method. */

void Ext2FileOpener::read_directory(fileid_t fileid, file_t *writer)
{
    /* Initialise directory reading state: opener, writer, entry, offset. */

    struct Ext2FileOpenerDir dir = {this, writer, NULL, 0};

    /* Call the handler function for each directory entry. */

    _ops->directory_iterate((ext2_ino_t) fileid, read_directory_entry, &dir);
}

/* Callback method for directory entry output. */

int Ext2FileOpener::write_directory_entry(struct Ext2FileOpenerDir *dir)
{
    struct ext2_inode inode;

    /* Obtain the inode details for metadata. */

    if (_ops->read_inode(dir->entry->inode, &inode))
        return DIRENT_ABORT;

    /* Align the size of the entry to avoid problems on architectures which
       require aligned accesses and where the compiler needs to assume an
       aligned structure. */

    offset_t namelen = ext2fs_dirent_name_len(dir->entry);
    offset_t reclen = pad_align(DIRENT_CORE_SIZE + namelen);

    /* Construct a directory entry structure of the calculated size. */

    char buffer[reclen];
    struct dirent *dirent = (struct dirent *) buffer;

    dirent->d_ino = dir->entry->inode;
    dirent->d_off = dir->offset;
    dirent->d_reclen = reclen;
    dirent->d_type = convert_file_type(ext2fs_dirent_file_type(dir->entry));

    /* Copy the name, padding the memory after it to the alignment boundary. */

    memcpy(dirent->d_name, dir->entry->name, namelen);
    memset(dirent->d_name + namelen, 0, reclen - namelen);

    /* Write the structure to the pipe. */

    offset_t nwritten = 0;

    while (nwritten < reclen)
        nwritten += client_write(dir->writer, (const void *) (dirent + nwritten), reclen - nwritten);

    return 0;
}

/* Return a file identifier for the given 'path'. */

long Ext2FileOpener::get_fileid(const char *path, flags_t flags, fileid_t *fileid)
{
    ext2_ino_t ino;
    const char *remaining;
    long err = _ops->find_file(path, &ino, &remaining);

    /* Return an existing file. */

    if (!err)
    {
        *fileid = (fileid_t) ino;
        return L4_EOK;
    }

    *fileid = FILEID_INVALID;

    /* Create a missing file if possible. */

    if ((flags & O_CREAT) && (flags & ~O_DIRECTORY))
    {
        /* Determine whether only the leafname is left of the path, with
           the inode number referring to the parent directory. */

        if (path_is_leafname(remaining))
        {
            err = _ops->create_file(ino, remaining, _user, &ino);
            if (err)
                return err;

            *fileid = (fileid_t) ino;
            return L4_EOK;
        }
    }

    return -L4_ENOENT;
}

/* Return a new accessor for 'fileid'. */

long Ext2FileOpener::make_accessor(const char *path, flags_t flags, fileid_t fileid, Accessor **accessor)
{
    (void) path; (void) flags;

    ext2_file_t file;
    long err = _ops->open_file((ext2_ino_t) fileid, &file);

    if (err)
        return err;

    *accessor = new Ext2FileAccessor(_ops, file, fileid);
    return L4_EOK;
}

// vim: tabstop=4 expandtab shiftwidth=4
